"""ORM client for a specific cell in an inventory in SciNote."""

import logging

from ..client.api.inventory_cell_client import InventoryCellClient
from ..client.models.inventory_cell import UpdateInventoryCell

logger = logging.getLogger(__name__)


class Cell:
    """ORM client for a specific cell in an inventory in SciNote."""

    def __init__(self, cell_id: int, attributes, client: InventoryCellClient):
        self.cell_id = cell_id
        self.attributes = attributes
        self.__client = client

    def value(self):
        """Get the value of the cell."""

        logger.debug(f'Getting value for cell {self.attributes}')
        match self.attributes.value_type:
            case 'text':
                return self.attributes.value.text
            case 'number':
                return self.attributes.value.data
            case 'list':
                return self.attributes.value.inventory_list_item_name
            case 'date_time':
                return self.attributes.value.date_time
            case 'date':
                return self.attributes.value.date
            case _:
                raise ValueError(
                    f'Unknown value type {self.attributes.value_type}'
                )

    def update_value(self, value: str):
        """Update the value stored in cell."""
        logger.debug(f'Updating value for cell {self.attributes}')
        try:
            result = self.__client.update_cell(
                UpdateInventoryCell(
                    id=self.cell_id,
                    value=value,
                    column_id=self.attributes.column_id,
                )
            )
            if result:
                self.attributes = result.attributes
        except Exception as e:
            logger.error(f'Failed to update cell {self.cell_id}: {e}')
            raise



"""Model representing a team in SciNote."""

from pydantic import BaseModel, ConfigDict


class Attributes(BaseModel):
    name: str
    data_type: str


class InventoryColumn(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    type: str
    attributes: Attributes



"""SciNote REST API client for inventory column list items."""

import requests
import logging

from typing import List
from pydantic import ValidationError

from ..models.inventory_column_list_item import (
    InventoryColumnListItem,
    CreateInventoryColumnListItem,
)

logger = logging.getLogger(__name__)


class InventoryColumnListItemClient:
    """SciNote REST API client for inventory column list items."""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        team_id: str,
        inventory_id: str,
        column_id: str,
    ):
        """Initialize the client with the base URL and token."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.inventory_id = inventory_id
        self.column_id = column_id
        self.headers = {
            'Api-Key': api_key,
        }

    def get_list_items(self) -> List[InventoryColumnListItem]:
        """Get a list of inventory column list items."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/columns/{self.column_id}/list_items'
        )

        result = []

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        while url:
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
            visited.add(url)
            logger.debug(f'Get columns response: {response.json()}')
            try:
                [
                    result.append(InventoryColumnListItem(**column))
                    for column in response.json()['data']
                ]
            except ValidationError as e:
                raise ValueError(f'Invalid response from SciNote: {e}')

            url = response.json()['links'].get('next', None)

            # Check we haven't visited this url before
            if url in visited:
                break

        return result

    def create_list_item(
        self, item: CreateInventoryColumnListItem
    ) -> InventoryColumnListItem:
        """Create a new list item."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/columns/{self.column_id}/list_items'
        )
        response = requests.post(
            url,
            json={
                'data': {
                    'type': item.type,
                    'attributes': {
                        'data': item.data,
                    },
                },
            },
            headers=self.headers,
        )
        response.raise_for_status()
        try:
            return InventoryColumnListItem(**response.json()['data'])
        except ValidationError as e:
            logger.error(f'Invalid response from SciNote: {e}')
            raise



"""Maps all of the inventories for a SciNote team."""

import logging

from .inventory import Inventory
from ..client.api.inventory_client import InventoryClient

logger = logging.getLogger(__name__)


class Inventories:
    """Maps all of the inventories for a SciNote team."""

    def __init__(self, client: InventoryClient):
        self.__client = client
        self.__inventories = {}

    def inventory(self, name: str) -> Inventory:
        """Get the inventory by name."""

        # TODO: Handle a new inventory being added after the initial load.
        # This is an edge case but we could do something with webhooks
        # to invalidate the cache.
        if self.__inventories.get(name) is None:
            self.__load_inventories()

        logger.debug(f'Checking again for inventory {name}')
        if self.__inventories.get(name) is None:
            raise ValueError(f'Inventory {name} not found')

        return self.__inventories.get(name)

    def inventories(self) -> list[Inventory]:
        """Get all of the inventories."""
        if not self.__inventories:
            self.__load_inventories()
        return [value for value in self.__inventories.values()]

    def __load_inventories(self):
        logger.debug('Loading all inventories')

        scinote_inventories = self.__client.get_inventories()
        for inventory in scinote_inventories:
            inventory_name = inventory.attributes.name.lower().replace(' ', '_')
            logger.debug(f'Adding inventory {inventory_name}')
            self.__inventories[inventory_name] = Inventory(
                inventory_name,
                self.__client.column_client(inventory.id),
                self.__client.item_client(inventory.id),
            )



"""SciNote rest api client for inventories."""

import requests
import logging

from typing import List
from pydantic import ValidationError

from .inventory_column_client import InventoryColumnClient
from .inventory_item_client import InventoryItemClient
from ..models.inventories import Inventory

logger = logging.getLogger(__name__)


class InventoryClient:
    """Client for interacting with the SciNote REST API."""

    def __init__(self, base_url: str, api_key: str, team_id: int):
        """Initialize the client."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.headers = {
            'Api-Key': api_key,
        }

    def column_client(self, inventory_id: int) -> InventoryColumnClient:
        return InventoryColumnClient(
            self.base_url, self.api_key, self.team_id, inventory_id
        )

    def item_client(self, inventory_id: int) -> InventoryItemClient:
        return InventoryItemClient(
            self.base_url, self.api_key, self.team_id, inventory_id
        )

    # Inventory related methods
    def get_inventories(self) -> List[Inventory]:
        """Get all inventories from SciNote."""
        logger.debug('Getting all inventories')
        # Need to loop through all pages to get all of the inventories
        url = f'{self.base_url}/api/v1/teams/{self.team_id}/inventories'
        result = []

        # Keep track of visited urls to avoid infinite loops
        visited = set(url)
        while url:
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
            visited.add(url)
            try:
                [
                    result.append(Inventory(**inventory))
                    for inventory in response.json()['data']
                ]
            except ValidationError as e:
                raise ValueError(f'Invalid response from SciNote: {e}')

            url = response.json()['links'].get('next', None)

            # Check we haven't visited this url before
            if url in visited:
                break

        logger.debug(f'Inventories retrieved: {result}')
        return result

    def create_inventory(self, name: str) -> Inventory:
        """Create a new inventory in SciNote."""
        logger.debug('Creating inventory %s', name)
        url = f'{self.base_url}/api/v1/teams/{self.team_id}/inventories'
        response = requests.post(
            url,
            headers=self.headers,
            json={
                'data': {
                    'type': 'inventories',
                    'attributes': {'name': name},
                }
            },
        )
        response.raise_for_status()
        try:
            return Inventory(**(response.json()['data']))
        except ValueError as e:
            logger.error(f'Invalid response: {e}')
            raise



"""ORM client for an specific inventory in SciNote."""

import logging

from .items import Item
from ..client.api.inventory_column_client import InventoryColumnClient
from ..client.api.inventory_item_client import InventoryItemClient
from ..client.models.inventory_cell import CreateInventoryCell

logger = logging.getLogger(__name__)


class Inventory:
    """ORM client for a specific inventory in SciNote."""

    def __init__(
        self,
        name: str,
        column_client: InventoryColumnClient,
        item_client: InventoryItemClient,
    ):
        self.column_client = column_client
        self.item_client = item_client
        self.name = name
        self.__columns = {}
        self.__item_list = []

    @staticmethod
    def load_columns(func):
        """Load the columns for this inventory."""

        def wrapper(cls, *args, **kwargs):
            if not cls.__columns:
                columns = cls.column_client.get_columns()
                for column in columns:
                    name = column.attributes.name.lower().replace(' ', '_')
                    cls.__columns[name] = column
            return func(cls, *args, **kwargs)

        return wrapper

    @load_columns
    def items(self) -> list[Item]:
        """Get the items for this inventory."""

        # If we've already loaded the items, return them.
        if self.__item_list:
            return self.__item_list

        # We need all of the columns to pass to each item. It would be good
        # to time this out somehow so we get the columns every so often.
        if not self.__columns:
            self.__load_columns()

        items = self.item_client.get_items()
        for item in items:
            self.__item_list.append(
                Item(
                    item.id,
                    item.attributes.name,
                    self.item_client,
                    self.__columns,
                    item.inventory_cells,
                )
            )

        return self.__item_list

    @load_columns
    def columns(self):
        """Get the columns for this inventory."""
        return [column for column in self.__columns.values()]

    @load_columns
    def create_item(self, name: str, **kwargs) -> Item:
        """Create a new item in this inventory."""
        cells = []
        for key, value in kwargs.items():
            if key not in self.__columns:
                raise ValueError(f'Column {key} does not exist in inventory.')

            column = self.__columns[key]
            cells.append(CreateInventoryCell(value=value, column_id=column.id))

        item = self.item_client.create_item(name, cells)

        new_item = Item(
            item.id,
            item.attributes.name,
            self.item_client,
            self.__columns,
            item.inventory_cells,
        )
        self.__item_list.append(new_item)
        return new_item



"""SciNote REST API client for inventory items."""

import requests

from typing import List
from pydantic import ValidationError

from .inventory_cell_client import InventoryCellClient
from .inventory_item_parent_client import InventoryParentClient
from .inventory_item_child_client import InventoryChildClient
from ..models.inventory_cell import InventoryCell, CreateInventoryCell
from ..models.inventory_item import InventoryItem


class InventoryItemClient:
    """Client for interacting with the SciNote REST API."""

    def __init__(
        self, base_url: str, api_key: str, team_id: int, inventory_id: int
    ):
        """Initialize the client."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.inventory_id = inventory_id
        self.headers = {
            'Api-Key': api_key,
        }
        self.params = {
            'include': 'inventory_cells',
        }

    def cell_client(self, item: int) -> InventoryCellClient:
        return InventoryCellClient(
            self.base_url, self.api_key, self.team_id, self.inventory_id, item
        )

    def parent_client(self, item: int) -> InventoryParentClient:
        return InventoryParentClient(
            self.base_url, self.api_key, self.team_id, self.inventory_id, item
        )

    def child_client(self, item: int) -> InventoryChildClient:
        return InventoryChildClient(
            self.base_url, self.api_key, self.team_id, self.inventory_id, item
        )

    def get_items(self) -> List[InventoryItem]:
        """Get all inventory items from SciNote."""

        # Need to loop through all pages to get all of the items
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items'
        )
        result = []
        inventory_cells = {}

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        while url:
            response = requests.get(
                url, headers=self.headers, params=self.params
            )
            response.raise_for_status()
            visited.add(url)
            try:
                [
                    result.append(InventoryItem(**item))
                    for item in response.json()['data']
                ]
            except ValidationError as e:
                raise ValueError(f'Invalid response from SciNote: {e}')

            # Read of all the cell responses as well so we can marry them
            # together later.
            try:
                [
                    inventory_cells.update({cell['id']: InventoryCell(**cell)})
                    for cell in response.json()['included']
                ]
            except ValidationError as e:
                raise ValueError(f'Invalid response from SciNote: {e}')

            url = response.json()['links'].get('next', None)

            # Check we haven't visited this url before
            if url in visited:
                break

        # If we have inventory cells, we need to attach them to the items.
        for item in result:
            item.inventory_cells = [
                inventory_cells[cell['id']]
                for cell in item.relationships.inventory_cells['data']
            ]

        return result

    def create_item(
        self, name: str, cells: List[CreateInventoryCell] = None
    ) -> str:
        """Create a new inventory item in SciNote."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items'
        )

        included = []
        if cells:
            for cell in cells:
                included.append(
                    {
                        'type': cell.type,
                        'attributes': {
                            'value': cell.value,
                            'column_id': cell.column_id,
                        },
                    }
                )

        response = requests.post(
            url,
            headers=self.headers,
            json={
                'data': {
                    'type': 'inventory_items',
                    'attributes': {
                        'name': name,
                    },
                },
                'included': included,
            },
        )
        response.raise_for_status()
        item = InventoryItem(**response.json()['data'])

        if 'included' in response.json():
            for cell in response.json()['included']:
                item.inventory_cells.append(InventoryCell(**cell))

        return item



"""SciNote rest api client."""

import requests
import os

from typing import List
from pydantic import ValidationError

from .inventory_client import InventoryClient
from ..models.teams import Team


def CreateClient() -> 'TeamsClient':
    url = os.getenv('SCINOTE_BASE_URL')
    api_key = os.getenv('SCINOTE_API_KEY')

    assert url, 'SCINOTE_BASE_URL environment variable is required.'
    assert api_key, 'SCINOTE_API_KEY environment variable is required.'

    return TeamsClient(url, api_key)


class TeamsClient:
    """Client for interacting with the SciNote REST API."""

    def __init__(self, base_url: str, api_key: str):
        """Initialize the client with the base URL and API key."""
        self.base_url = base_url
        self.api_key = api_key
        self.headers = {
            'Api-Key': api_key,
        }

    def inventory_client(self, team_id: int) -> InventoryClient:
        return InventoryClient(self.base_url, self.api_key, team_id)

    # Teams related methods
    def get_teams(self) -> List[Team]:
        """Get all teams from SciNote."""
        # Need to loop through all pages to get all of the inventories
        url = f'{self.base_url}/api/v1/teams'
        result = []

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        while url:
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
            visited.add(url)
            try:
                [
                    result.append(Team(**team))
                    for team in response.json()['data']
                ]
            except ValidationError as e:
                raise ValueError(f'Invalid response from SciNote: {e}')

            url = response.json()['links'].get('next', None)

            # Check we haven't visited this url before
            if url in visited:
                break

        return result



"""Model representing an inventory column list item in SciNote."""

from datetime import datetime

from pydantic import BaseModel, ConfigDict, Field


class Attributes(BaseModel):
    data: str
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)


class InventoryColumnListItem(BaseModel):
    model_config = ConfigDict(extra='ignore')

    id: str
    type: str
    attributes: Attributes


class CreateInventoryColumnListItem(BaseModel):
    """Model represents a new inventory column list item in SciNote."""

    type: str = 'inventory_list_items'
    data: str



"""SciNote rest api client for inventory columns."""

import requests

from typing import List
from pydantic import ValidationError

from ..models.inventory_parent import InventoryParent


class InventoryParentClient:
    """Client for interacting with the SciNote REST API."""

    def __init__(
        self,
        base_url: str,
        api_key: str,
        team_id: int,
        inventory_id: int,
        item_id: int,
    ):
        """Initialize the client."""
        self.base_url = base_url
        self.api_key = api_key
        self.team_id = team_id
        self.inventory_id = inventory_id
        self.item_id = item_id
        self.headers = {
            'Api-Key': api_key,
        }

    def get_parents(self) -> List[InventoryParent]:
        """Get all inventory parents for the given item from SciNote."""

        # Need to loop through all pages to get all of the columns
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items/{self.item_id}/parent_relationships'
        )
        result = []

        # Keep track of visited urls to avoid infinite loops
        visited = set()
        while url:
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
            visited.add(url)
            try:
                [
                    result.append(InventoryParent(**parent))
                    for parent in response.json()['data']
                ]
            except ValidationError as e:
                raise ValueError(f'Invalid response from SciNote: {e}')

            url = response.json()['links'].get('next', None)

            # Check we haven't visited this url before
            if url in visited:
                break

        return result

    def create_parent(self, parent_id: str) -> str:
        """Create a new parent for the given item in SciNote."""
        url = (
            f'{self.base_url}/api/v1/teams/{self.team_id}/inventories/'
            f'{self.inventory_id}/items/{self.item_id}/parent_relationships'
        )

        response = requests.post(
            url,
            headers=self.headers,
            json={
                'data': {
                    'type': 'inventory_item_relationships',
                    'attributes': {
                        'parent_id': parent_id,
                    },
                }
            },
        )
        response.raise_for_status()
        return response.json()['data']['id']
